"""This module contains the context managers for the Translations app."""

from django.db import models

import translations.models
from translations.languages import _get_default_language, \
    _get_translate_language
from translations.utils import _get_relations_hierarchy, _get_purview, \
    _get_translations


__docformat__ = 'restructuredtext'


class Context:
    """A context manager which provides custom translation functionalities."""

    def __init__(self, entity, *relations):
        """Initialize a `Context` with an entity and some relations of it."""
        hierarchy = _get_relations_hierarchy(*relations)
        self.mapping, self.query = _get_purview(entity, hierarchy)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _get_changed_fields(self):
        r"""
        Yield the info about the changed fields in the `Context`\ 's `purview`.
        """
        for (ct_id, objs) in self.mapping.items():
            for (obj_id, obj) in objs.items():
                for field in type(obj)._get_translatable_fields_names():
                    text = getattr(obj, field, None)
                    default = obj._default_translatable_fields.get(field, None)
                    if text and text != default:
                        yield ({
                            'content_type_id': ct_id,
                            'object_id': obj_id,
                            'field': field,
                        }, text)

    def create(self, lang=None):
        r"""
        Create the translations of the `Context`\ 's `purview` in a language.
        """
        lang = _get_translate_language(lang)
        if lang != _get_default_language():
            _translations = [
                translations.models.Translation(
                    language=lang, text=text, **address
                ) for address, text in self._get_changed_fields()
            ]
            translations.models.Translation.objects.bulk_create(_translations)

    def read(self, lang=None):
        r"""
        Read the translations of the `Context`\ 's `purview` in a language.
        """
        lang = _get_translate_language(lang)
        if lang != _get_default_language():
            _translations = _get_translations(self.query, lang)
            for translation in _translations:
                ct_id = translation.content_type.id
                obj_id = translation.object_id
                field = translation.field
                text = translation.text
                obj = self.mapping[ct_id][obj_id]
                if field in type(obj)._get_translatable_fields_names():
                    setattr(obj, field, text)
        else:
            self.reset()

    def update(self, lang=None):
        r"""
        Update the translations of the `Context`\ 's `purview` in a language.
        """
        lang = _get_translate_language(lang)
        if lang != _get_default_language():
            query = models.Q()
            _translations = []
            for address, text in self._get_changed_fields():
                query |= models.Q(**address)
                _translations.append(
                    translations.models.Translation(
                        language=lang, text=text, **address
                    )
                )
            _get_translations(query, lang).delete()
            translations.models.Translation.objects.bulk_create(_translations)

    def delete(self, lang=None):
        r"""
        Delete the translations of the `Context`\ 's `purview` in a language.
        """
        lang = _get_translate_language(lang)
        if lang != _get_default_language():
            _get_translations(self.query, lang).delete()

    def reset(self):
        r"""
        Reset the translations of the `Context`\ 's `purview` to
        the `default language`.
        """
        for (ct_id, objs) in self.mapping.items():
            for (obj_id, obj) in objs.items():
                for (field, value) in obj._default_translatable_fields.items():
                    setattr(obj, field, value)



"""This module contains the form utilities for the Translations app."""

from django import forms

from translations.models import Translation
from translations.languages import _get_translation_choices


__docformat__ = 'restructuredtext'


def generate_translation_form(translatable):
    r"""
    Return the `Translation` form based on a `Translatable` model and
    the `translation language`\ s.
    """
    fields = translatable._get_translatable_fields_choices()
    languages = _get_translation_choices()

    class TranslationForm(forms.ModelForm):
        field = forms.ChoiceField(choices=fields)
        language = forms.ChoiceField(choices=languages)

        class Meta:
            model = Translation
            fields = (
                'field',
                'language',
                'text',
            )

    return TranslationForm



"""This module contains the models for the Translations app."""

from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey, \
    GenericRelation
from django.utils.translation import ugettext_lazy as _

from translations.querysets import TranslatableQuerySet


__docformat__ = 'restructuredtext'


class Translation(models.Model):
    """The model which represents the translations."""

    content_type = models.ForeignKey(
        verbose_name=_('content type'),
        help_text=_('the content type of the object to translate'),
        to=ContentType,
        on_delete=models.CASCADE,
    )
    object_id = models.CharField(
        verbose_name=_('object id'),
        help_text=_('the id of the object to translate'),
        max_length=128,
    )
    content_object = GenericForeignKey(
        ct_field='content_type',
        fk_field='object_id',
    )
    field = models.CharField(
        verbose_name=_('field'),
        help_text=_('the field of the object to translate'),
        max_length=64,
    )
    language = models.CharField(
        verbose_name=_('language'),
        help_text=_('the language of the translation'),
        max_length=32,
        choices=settings.LANGUAGES,
    )
    text = models.TextField(
        verbose_name=_('text'),
        help_text=_('the text of the translation'),
    )

    def __str__(self):
        """Return the representation of the translation."""
        return '{source}: {translation}'.format(
            source=getattr(self.content_object, self.field),
            translation=self.text,
        )

    class Meta:
        unique_together = ('content_type', 'object_id', 'field', 'language',)
        verbose_name = _('translation')
        verbose_name_plural = _('translations')


class Translatable(models.Model):
    """An abstract model which provides custom translation functionalities."""
    objects = TranslatableQuerySet.as_manager()
    translations = GenericRelation(
        Translation,
        content_type_field='content_type',
        object_id_field='object_id',
        related_query_name='%(app_label)s_%(class)s',
    )

    class Meta:
        abstract = True

    class TranslatableMeta:
        """
        This class contains meta information about the translation
        of the model instances.
        """

        fields = None

    @classmethod
    def get_translatable_fields(cls):
        """Return the model’s translatable fields."""
        if not hasattr(cls, '_cached_translatable_fields'):
            if cls.TranslatableMeta.fields is None:
                fields = []
                for field in cls._meta.get_fields():
                    if isinstance(
                                field,
                                (models.CharField, models.TextField,)
                            ) and not isinstance(
                                field,
                                models.EmailField
                            ) and not (
                                hasattr(field, 'choices') and field.choices
                            ):
                        fields.append(field)
            else:
                fields = [
                    cls._meta.get_field(field_name)
                    for field_name in cls.TranslatableMeta.fields
                ]
            cls._cached_translatable_fields = fields
        return cls._cached_translatable_fields

    @classmethod
    def _get_translatable_fields_names(cls):
        """Return the names of the model's translatable fields."""
        if not hasattr(cls, '_cached_translatable_fields_names'):
            cls._cached_translatable_fields_names = [
                field.name for field in cls.get_translatable_fields()
            ]
        return cls._cached_translatable_fields_names

    @classmethod
    def _get_translatable_fields_choices(cls):
        """Return the choices of the model's translatable fields."""
        choices = [
            (None, '---------'),
        ]

        for field in cls.get_translatable_fields():
            choice = (field.name, field.verbose_name.title())
            choices.append(choice)

        return choices



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class TranslationsConfig(AppConfig):
    name = 'translations'
    verbose_name = _('translations')

    def ready(self):
        try:
            # cache all content types at the start
            from django.contrib.contenttypes.models import ContentType
            models = [ct.model_class() for ct in ContentType.objects.all()]
            ContentType.objects.get_for_models(*models)
        except Exception:
            pass



# Generated by Django 2.1.1 on 2018-09-20 12:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('translations', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='translation',
            name='language',
            field=models.CharField(choices=[('en', 'English'), ('en-gb', 'English (Great Britain)'), ('de', 'German'), ('tr', 'Turkish')], help_text='the language of the translation', max_length=32, verbose_name='language'),
        ),
    ]



"""This module contains the querysets for the Translations app."""

from django.db.models import query

from translations.languages import _get_default_language, \
    _get_translate_language, _get_probe_language
from translations.query import _fetch_translations_query_getter
from translations.context import Context


__docformat__ = 'restructuredtext'


class TranslatableQuerySet(query.QuerySet):
    """A queryset which provides custom translation functionalities."""

    def __init__(self, *args, **kwargs):
        """Initialize a `TranslatableQuerySet` with `QuerySet` arguments."""
        super(TranslatableQuerySet, self).__init__(*args, **kwargs)
        self._trans_lang = _get_default_language()
        self._trans_prob = _get_default_language()
        self._trans_rels = ()
        self._trans_cache = False

    def _chain(self, **kwargs):
        """Return a copy of the current `TranslatableQuerySet`."""
        clone = super(TranslatableQuerySet, self)._chain(**kwargs)

        # default values for all
        clone._trans_lang = getattr(self, '_trans_lang')
        clone._trans_prob = getattr(self, '_trans_prob')
        clone._trans_rels = getattr(self, '_trans_rels')

        # reset cache on chaining
        clone._trans_cache = False

        return clone

    def _fetch_all(self):
        """Evaluate the `TranslatableQuerySet`."""
        super(TranslatableQuerySet, self)._fetch_all()

        if self._trans_lang == _get_default_language():
            return

        if self._iterable_class is not query.ModelIterable:
            raise TypeError(
                'Translations does not support custom iteration (yet). ' +
                'e.g. values, values_list, etc. ' +
                'If necessary you can `decipher` and then do it.'
            )

        if not self._trans_cache:
            with Context(self._result_cache, *self._trans_rels) \
                    as context:
                context.read(self._trans_lang)
            self._trans_cache = True

    def translate(self, lang=None):
        """Translate the `TranslatableQuerySet` in a language."""
        clone = self.all()
        clone._trans_lang = _get_translate_language(lang)
        return clone

    def translate_related(self, *fields):
        """Translate some relations of the `TranslatableQuerySet`."""
        clone = self.all()
        clone._trans_rels = () if fields == (None,) else fields
        return clone

    def probe(self, lang=None):
        """Probe the `TranslatableQuerySet` in some language(s)."""
        clone = self.all()
        clone._trans_prob = _get_probe_language(lang)
        return clone

    def filter(self, *args, **kwargs):
        """Filter the `TranslatableQuerySet`."""
        if not (args or kwargs):
            return super(TranslatableQuerySet, self).filter()

        query = _fetch_translations_query_getter(
            self.model,
            self._trans_prob
        )(*args, **kwargs)
        return super(TranslatableQuerySet, self).filter(query)

    def exclude(self, *args, **kwargs):
        """Exclude the `TranslatableQuerySet`."""
        if not (args or kwargs):
            return super(TranslatableQuerySet, self).exclude()

        query = _fetch_translations_query_getter(
            self.model,
            self._trans_prob
        )(*args, **kwargs)
        return super(TranslatableQuerySet, self).exclude(query)



"""This module contains the query utilities for the Translations app."""

import copy

from django.db.models import Q
from django.db.models.constants import LOOKUP_SEP

from translations.languages import _get_default_language, _get_probe_language
from translations.utils import _get_dissected_lookup


__docformat__ = 'restructuredtext'


def _fetch_translations_query_getter(model, lang):
    """
    Return the translations query getter specialized for a model and some
    language(s).
    """
    default = _get_default_language()

    def _get_translations_query(*args, **kwargs):
        connector = kwargs.pop('_connector', None)
        negated = kwargs.pop('_negated', False)

        children = list(args) + sorted(kwargs.items())

        for index, child in enumerate(children):
            if isinstance(child, tuple):
                dissected = _get_dissected_lookup(model, child[0])
                if dissected['translatable']:
                    query_default = False
                    query_languages = None
                    if isinstance(lang, (list, tuple)):
                        query_languages = []
                        for x in lang:
                            if x == default:
                                query_default = True
                            else:
                                query_languages.append(x)
                    else:
                        if lang == default:
                            query_default = True
                            query_languages = None
                        else:
                            query_default = False
                            query_languages = lang

                    q = Q()

                    if query_default:
                        q |= Q(**{child[0]: child[1]})

                    if query_languages:
                        relation = LOOKUP_SEP.join(
                            dissected['relation'] + ['translations'])
                        field_supp = (LOOKUP_SEP + dissected['supplement']) \
                            if dissected['supplement'] else ''
                        lang_supp = (LOOKUP_SEP + 'in') \
                            if isinstance(query_languages, (list, tuple)) \
                            else ''

                        q |= Q(**{
                            '{}__field'.format(relation):
                                dissected['field'],
                            '{}__text{}'.format(relation, field_supp):
                                child[1],
                            '{}__language{}'.format(relation, lang_supp):
                                query_languages
                        })
                else:
                    q = Q(**{child[0]: child[1]})
            elif isinstance(child, TQ):
                if child.lang:
                    getter = _fetch_translations_query_getter(
                        model,
                        child.lang
                    )
                    q = getter(
                        *child.children,
                        _connector=child.connector,
                        _negated=child.negated
                    )
                else:
                    q = _get_translations_query(
                        *child.children,
                        _connector=child.connector,
                        _negated=child.negated
                    )
            elif isinstance(child, Q):
                q = _get_translations_query(
                    *child.children,
                    _connector=child.connector,
                    _negated=child.negated
                )
            children[index] = q

        query = Q(*children, _connector=connector, _negated=negated)

        return query

    return _get_translations_query


class TQ(Q):
    """
    Encapsulate translation queries as objects that can then be combined
    logically (using `&` and `|`).
    """

    def __init__(self, *args, **kwargs):
        """Initialize a `TQ` with `Q` arguments."""
        super(TQ, self).__init__(*args, **kwargs)
        self.lang = None

    def __deepcopy__(self, memodict):
        """Return a copy of the `TQ` object."""
        obj = super(TQ, self).__deepcopy__(memodict)
        obj.lang = self.lang
        return obj

    def __call__(self, lang=None):
        """Specialize the `TQ` for some language(s)."""
        obj = copy.deepcopy(self)
        obj.lang = _get_probe_language(lang)
        return obj

    def _combine(self, other, conn):
        """Return the result of logical combination with another `Q` object."""
        if not isinstance(other, Q):
            raise TypeError(other)

        # If the other Q() is empty, ignore it and just use `self`.
        if not other:
            return copy.deepcopy(self)
        # Or if this Q is empty, ignore it and just use `other`.
        elif not self:
            return copy.deepcopy(other)

        obj = Q(self, other, _connector=conn)
        return obj
